"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
function optimisticUi(params) {
    const { commitTransaction, onChangePendingTransactionCount, onError, onRerenderKind, optimisticUIEnabled } = params;
    let pendingTransactions = [];
    return {
        pendingTransactionCount: () => pendingTransactions.length,
        update: function (kind, id, delta) {
            const transaction = { kind, id, delta };
            pendingTransactions.push(transaction);
            onChangePendingTransactionCount();
            // Rerender immediately to show optimistic update
            onRerenderKind(kind, id);
            commitTransaction(kind, id, delta)
                .then(function () {
                pendingTransactions = pendingTransactions.filter(t => t !== transaction);
                onChangePendingTransactionCount();
                // We have removed the pending transaction asynchronously,
                // and the server may have already added changes on top,
                // which the UI does not show yet.
                onRerenderKind(kind, id);
            })
                .catch(onError);
        },
        withPendingTransactions: function (data) {
            if (!optimisticUIEnabled())
                return data;
            const result = {};
            Object.keys(data).forEach(k => {
                result[k] = Object.assign({}, data[k]);
            });
            // Apply pending transactions on top of the canon version
            pendingTransactions.forEach(function (transaction) {
                const prev = result[transaction.kind][transaction.id];
                if (!prev)
                    throw new Error(`withPendingTransactions ${transaction.kind} ${transaction.id} called before the data has been retrieved`);
                result[transaction.kind][transaction.id] = {
                    revision: "optimistic",
                    value: transaction.delta(prev.value)
                };
            });
            return result;
        }
    };
}
exports.optimisticUi = optimisticUi;
//# sourceMappingURL=optimisticUi.js.map