"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const throttled_1 = require("./throttled");
const defaultRetries = 15;
const sleep = (ms) => new Promise(resolve => setTimeout(resolve, ms));
function makePush(params) {
    const { canonData, onError, onNetPush, onUpdate, shouldCrashWrites } = params;
    let pushCounter = 0;
    function attemptPush(params) {
        return __awaiter(this, void 0, void 0, function* () {
            if (shouldCrashWrites()) {
                yield sleep(2000);
                throw new Error("Simulate write failure");
            }
            const { kind, id, deltas, previousConflictOnRevision, retriesRemaining } = params;
            const curr = canonData[kind][id];
            if (!curr)
                throw new Error(`Pushing to ${kind} while the canon data is empty is not yet implemented. Proper implementation turns on observing here, waits for the data and tries to push.`);
            const value = deltas.reduce((v, delta) => delta(v), curr.value);
            const pushId = `push-${pushCounter++}`;
            const result = yield onNetPush({
                kind,
                id,
                pushId,
                lastSeenRevision: curr.revision,
                value
            });
            if (result === "conflict") {
                if (previousConflictOnRevision &&
                    curr.revision === previousConflictOnRevision) {
                    throw new Error(`The revision ${curr.revision} is not changing and the server keeps responding with a conflict. Check whether the server is correctly incrementing the revision and whether the client is subscribed to updates on ${kind} ${id}.`);
                }
                // Server and the connection are healthy, so reset the retries
                // We can resolve conflicts infinitely, it’s the internal errors that
                // we need to limit retries.
                return yield attemptPush(Object.assign({}, params, { previousConflictOnRevision: curr.revision, retriesRemaining: defaultRetries }));
            }
            if (result === "internalError") {
                // As many other pieces of this code, the retry logic needs
                // cleaner refactoring
                if (retriesRemaining === 0)
                    throw new Error(`Write failed after all retries`);
                yield sleep(3000);
                return yield attemptPush(Object.assign({}, params, { retriesRemaining: retriesRemaining - 1 }));
            }
            if (typeof result === "object") {
                onUpdate(Object.assign({ kind, id }, result));
                return;
            }
            throw new Error(`Unexpected result ${result}`);
        });
    }
    const queuedTasks = {};
    function queuedTasksFor(kind, id) {
        queuedTasks[kind] = queuedTasks[kind] || {};
        queuedTasks[kind][id] = queuedTasks[kind][id] || [];
        return queuedTasks[kind][id];
    }
    function performPush(kind, id) {
        return __awaiter(this, void 0, void 0, function* () {
            const taskQueue = queuedTasksFor(kind, id);
            const tasks = taskQueue.slice();
            taskQueue.splice(0);
            try {
                yield attemptPush({
                    kind,
                    id,
                    deltas: tasks.map(t => t.delta),
                    retriesRemaining: defaultRetries
                });
            }
            catch (err) {
                onError(err);
            }
            tasks.forEach(t => t.resolve());
        });
    }
    const run = throttled_1.makeThrottled(performPush);
    return function (kind, id, delta) {
        return new Promise(resolve => {
            queuedTasksFor(kind, id).push({ delta, resolve });
            run(kind, id);
        });
    };
}
exports.makePush = makePush;
//# sourceMappingURL=push.js.map