"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const nanoevents_1 = __importDefault(require("nanoevents"));
const ws_1 = __importDefault(require("ws"));
const serialization_1 = require("../serialization");
const authQueue_1 = require("./authQueue");
const types_1 = require("./types");
function runWebsocketServer(params) {
    const { auth, onChangeData, onRequestData } = params;
    const serialization = params.serialization || serialization_1.jsonSerialization;
    const events = new nanoevents_1.default();
    const latestCopies = {};
    events.on("change", kv => {
        latestCopies[types_1.stringy(kv)] = kv.value;
    });
    const listeningTo = [];
    function getAndObserve(key, close, onData) {
        if (listeningTo.includes(types_1.stringy(key))) {
            const last = latestCopies[types_1.stringy(key)];
            if (last)
                onData(last);
            return;
        }
        listeningTo.push(types_1.stringy(key));
        events.on("change", eventKV => {
            if (types_1.stringy(key) === types_1.stringy(eventKV))
                onData(eventKV.value);
        });
        onRequestData(Object.assign({}, key, { close: close, send: value => events.emit("change", Object.assign({}, key, { value })) }));
    }
    const server = params._ws ||
        new ws_1.default.Server({
            port: params.port,
            server: params.server
        });
    server.on("connection", function (socket) {
        const authQueue = authQueue_1.makeAuthQueue({
            onTerminate: () => socket.terminate()
        });
        const send = (obj) => {
            if (socket.readyState !== 1)
                return;
            socket.send(JSON.stringify(obj));
        };
        const handlers = {
            ping: () => send({ action: "pong" }),
            auth: msg => {
                const { token } = msg;
                if (!auth)
                    throw new Error("Unexpected auth message");
                authQueue.newAuthIncoming(auth
                    .parseToken(token, { close: () => socket.terminate() })
                    .catch(() => {
                    socket.terminate();
                    return undefined;
                }));
            },
            push: (msg) => __awaiter(this, void 0, void 0, function* () {
                const { kind, id, lastSeenRevision, value } = msg;
                if (auth) {
                    const details = yield authQueue.details();
                    if (!auth.canWrite({ auth: details, kind, id })) {
                        send({
                            action: "clientError",
                            message: `Tried to write ${kind} ${id} without permission`
                        });
                        return;
                    }
                }
                onChangeData({
                    kind,
                    lastSeenRevision,
                    id,
                    value: serialization.decode(value),
                    close: () => socket.terminate()
                })
                    .then(function (result) {
                    if (result === "conflict")
                        send({
                            action: "pushResult",
                            pushId: msg.pushId,
                            result: result
                        });
                    else
                        send({
                            action: "pushResult",
                            pushId: msg.pushId,
                            result: "success",
                            newRevision: result.newRevision,
                            newValue: serialization.encode(result.newValue)
                        });
                })
                    .catch(function () {
                    send({
                        action: "pushResult",
                        pushId: msg.pushId,
                        result: "internalError"
                    });
                });
            }),
            subscribe: msg => {
                const { id, kind } = msg;
                getAndObserve({ kind, id }, () => socket.terminate(), (v) => __awaiter(this, void 0, void 0, function* () {
                    if (auth) {
                        const details = yield authQueue.details();
                        if (!auth.canRead({ auth: details, kind, id })) {
                            send({
                                action: "clientError",
                                message: `Tried to read ${kind} ${id} without permission`
                            });
                            return;
                        }
                    }
                    send({
                        action: "update",
                        id,
                        kind,
                        revision: v.revision,
                        value: serialization.encode(v.value)
                    });
                }));
            }
        };
        const handleMsg = (msg) => {
            const action = msg.action;
            const handler = handlers[action];
            if (!handler)
                return console.warn("Unrecognized message", msg);
            handler(msg);
        };
        socket.on("message", function incoming(data) {
            function attempt(f) {
                try {
                    return f();
                }
                catch (err) {
                    return undefined;
                }
            }
            const msg = attempt(() => JSON.parse(data.toString()));
            if (!msg) {
                console.warn("Received invalid message format");
                socket.terminate();
                return;
            }
            handleMsg(msg);
        });
    });
}
exports.runWebsocketServer = runWebsocketServer;
//# sourceMappingURL=index.js.map